#!/usr/bin/env python3
"""Generate documented YAML files from documented dataclasses."""

import dataclasses
import functools
import logging
import textwrap
from typing import get_args

import yaml
from docstring_parser import parse as parse_docstring

from .typing import type_to_string


LOGGER = logging.getLogger(__name__)


class DataclassDocumenter:
    """Generate markdown and YAML documentation from documented dataclasses."""

    def __init__(self, datacls, name=None, width=120):
        """
        Args:
            datacls:
                The dataclass to document.

            name:
                The name to use for this class in the documentation. If None,
                the name of the dataclass will be used.

            width:
                The target output width for wrapped comments.
        """
        if not dataclasses.is_dataclass(datacls):
            raise ValueError("First argument is not a dataclass.")
        self.datacls = datacls
        self.name = datacls.__name__ if name is None else name
        self.width = int(width)

    @functools.cached_property
    def docstring(self):
        """The parsed docstring."""
        return parse_docstring(self.datacls.__doc__)

    @functools.cached_property
    def fields(self):
        """The fields of the dataclass."""
        return dataclasses.fields(self.datacls)

    @functools.cached_property
    def params(self):
        """The dict mapping parameter names to DocstrignParam values."""
        return {param.arg_name: param for param in self.docstring.params}

    def get_param_desc(self, param, default=None, warn=False):
        """
        Get the description of a parameter.

        Args:
            param:
                The parameter name.

            default:
                The default value to return if the parameter lacks a description.

            warn:
                If True, log a warning when the description is missing.

        Returns:
            The parameter description.
        """
        try:
            return self.params[param].description
        except KeyError:
            if warn:
                LOGGER.warning("%s is not documented in the docstring.", param)
            return default

    def get_param_type(self, param):
        """
        Get the type of a parameter.

        Args:
            param:
                The parameter name.

        Returns:
            The parameter type.
        """
        return self.datacls.__annotations__[param]

    def _wrap_yaml_comment(self, comment, indent):
        """
        Wrap a YAML comment.

        Args:
            comment:
                The comment to wrap.

            indent:
                The indentation for each line.

        Returns:
            The wrapped lines.
        """
        indent = f"{indent}# "
        for line in textwrap.wrap(
            comment, width=self.width, initial_indent=indent, subsequent_indent=indent
        ):
            yield line.rstrip()

    def _default_as_yaml(self, name, value, indent, commented=False):
        """
        Wrap YAML output for embedding in another YAML document.

        Args:
            name:
                The field name.

            value:
                The value.

            indent:
                The indentation for each line.

            commented:
                If True, comment fields.

        Returns:
            The wrapped YAML lines.
        """
        obj = {name: value}
        text = yaml.dump(obj)
        for line in text.splitlines():
            if commented:
                yield f"{indent}# {line}"
            else:
                yield f"{indent}{line}"

    def get_yaml_blocks(self, level=0, header=None, commented=False):
        """
        Get commented YAML input for the dataclass.

        Args:
            level:
                The indentation level.

            header:
                An optional header to emit as a comment at the start of the
                output.

            commented:
                If True, comment all fields.

        Returns:
            A generator over blocks of YAML.
        """
        indent_unit = "  "
        indent = indent_unit * level
        empty_line = ""
        if commented:
            field_indent = f"{indent}# "
        else:
            field_indent = indent
        if header is not None:
            yield from self._wrap_yaml_comment(header, indent)
            yield empty_line
        for field in self.fields:
            doc = self.get_param_desc(field.name, default="Undocumented.", warn=True)
            yield from self._wrap_yaml_comment(doc, indent)

            # Recursively document dataclasses.
            if dataclasses.is_dataclass(field.type):
                yield f"{field_indent}{field.name}:"
                dado = self.__class__(field.type)
                yield from dado.get_yaml_blocks(level=level + 1, commented=commented)
                continue

            meta = f"{indent}# Type: {type_to_string(field.type)}"
            if field.default is dataclasses.MISSING:
                if field.default_factory is dataclasses.MISSING:
                    yield f"{meta} [REQUIRED]"
                    yield f"{field_indent}{field.name}: ..."
                else:
                    yield f"{meta} [OPTIONAL]"
                    default = field.default_factory()
                    yield from self._default_as_yaml(
                        field.name, default, indent, commented=commented
                    )
            else:
                yield f"{meta} [OPTIONAL]"
                yield from self._default_as_yaml(
                    field.name, field.default, indent, commented=commented
                )
            emit_empty_line = True
            for arg in get_args(field.type):
                if dataclasses.is_dataclass(arg):
                    dado = self.__class__(arg)
                    yield from dado.get_yaml_blocks(
                        header=arg.__name__, level=level + 1, commented=commented
                    )
                    emit_empty_line = False

            if emit_empty_line:
                yield empty_line

    def get_yaml(self, level=0, commented=False):
        """
        Get commented YAML input for the dataclass.

        Args:
            level:
                Markdown header level.

            commented:
                If True, comment all lines.
        """
        header = self.name
        return "\n".join(
            self.get_yaml_blocks(level=level, header=header, commented=commented)
        )

    def get_markdown(self, level=0):
        """
        Get a markdown description of the dataclass that contains a commented
        example YAML input file.

        Args:
            level:
                Markdown header level.

        Returns:
            The markdown string.
        """
        level = max(level + 1, 1)
        header_prefix = "#" * level

        docstring = self.docstring
        cls_desc = (docstring.short_description, docstring.long_description)
        cls_desc = [desc for desc in cls_desc if desc]
        if cls_desc:
            cls_desc = "\n\n".join(cls_desc)
        return f"""{header_prefix} {self.name}

{cls_desc}

{header_prefix}# Input

~~~yaml
{self.get_yaml()}
~~~

"""
